@@ -21,7 +21,9 @@
 # Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
 import os
-from os import path
+from os import environ, path
+from locale import getpreferredencoding
+import signal
 import sys
 import gobject
 import ibus
@@ -70,13 +72,18 @@ CONV_MODE_WIDE_LATIN_2, \
 CONV_MODE_WIDE_LATIN_3, \
 CONV_MODE_PREDICTION = range(14)
 
+SEGMENT_DEFAULT         = 0
+SEGMENT_SINGLE          = 1 << 0
+SEGMENT_IMMEDIATE       = 1 << 1
+
 CLIPBOARD_RECONVERT = range(1)
 
-''' FIXME: currently configuration values are extracted by enviroment values.
-    It's better to load config.py(.in) in engine and setup instead and
-    I will move KASUMI_IMG_PATH.
-'''
-KASUMI_IMG_PATH = "/usr/share/pixmaps/kasumi.png"
+LINK_DICT_EMBEDDED, \
+LINK_DICT_SINGLE = range(2)
+
+IMPORTED_EMBEDDED_DICT_DIR = "imported_words_default.d"
+IMPORTED_EMBEDDED_DICT_PREFIX = "ibus__"
+IMPORTED_SINGLE_DICT_PREFIX = "imported_words_ibus__"
 
 KP_Table = {}
 for s in dir(keysyms):
@@ -96,6 +103,7 @@ class Engine(ibus.EngineBase):
     __setup_pid = 0
     __prefs = None
     __keybind = {}
+    __thumb = None
 
     def __init__(self, bus, object_path):
         super(Engine, self).__init__(bus, object_path)
@@ -105,8 +113,12 @@ class Engine(ibus.EngineBase):
         self.__context._set_encoding(anthy.ANTHY_UTF8_ENCODING)
 
         # init state
+        self.__idle_id = 0
         self.__input_mode = INPUT_MODE_HIRAGANA
+        self.__segment_mode = SEGMENT_DEFAULT
+        self.__dict_mode = 0
         self.__prop_dict = {}
+        self.__is_utf8 = (getpreferredencoding().lower() == "utf-8")
 
 #        self.__lookup_table = ibus.LookupTable(page_size=9, round=True)
         size = self.__prefs.get_value('common', 'page_size')
@@ -122,6 +134,11 @@ class Engine(ibus.EngineBase):
         mode = 'TypingMode.' + ['Romaji', 'Kana', 'ThumbShift'][mode]
         self.__typing_mode_activate(mode, ibus.PROP_STATE_CHECKED)
 
+        mode = self.__prefs.get_value('common', 'conversion_segment_mode')
+        mode = 'SegmentMode.' + ['Multi', 'Single',
+                                 'ImmediateMulti', 'ImmediateSingle'][mode]
+        self.__segment_mode_activate(mode, ibus.PROP_STATE_CHECKED)
+
         # use reset to init values
         self.__reset()
 
@@ -130,7 +147,6 @@ class Engine(ibus.EngineBase):
         self.__preedit_ja_string = jastring.JaString(Engine.__typing_mode)
         self.__convert_chars = u""
         self.__cursor_pos = 0
-        self.__need_update = False
         self.__convert_mode = CONV_MODE_OFF
         self.__segments = list()
         self.__lookup_table.clean()
@@ -138,6 +154,11 @@ class Engine(ibus.EngineBase):
         self._MM = 0
         self._SS = 0
         self._H = 0
+        self._RMM = 0
+        self._RSS = 0
+        if self.__idle_id != 0:
+            gobject.source_remove(self.__idle_id)
+            self.__idle_id = 0
 
     def __init_props(self):
         anthy_props = ibus.PropList()
@@ -199,20 +220,107 @@ class Engine(ibus.EngineBase):
         typing_mode_prop.set_sub_props(props)
         anthy_props.append(typing_mode_prop)
 
-        self.__set_dict_props(anthy_props)
+        self.__set_segment_mode_props(anthy_props)
+        self.__set_dict_mode_props(anthy_props)
+        self.__set_dict_config_props(anthy_props)
         anthy_props.append(ibus.Property(key=u"setup",
+                                         label=UN(_("Preferences - Anthy")),
+                                         icon=u"gtk-preferences",
                                          tooltip=UN(_("Configure Anthy"))))
 
         return anthy_props
 
-    def __set_dict_props(self, anthy_props):
-        path = self.__prefs.get_value('common', 'dict_admin_command')
-        if not os.path.exists(path[0]):
-            return
+    def __init_signal(self):
+        signal.signal(signal.SIGHUP, self.__signal_cb)
+        signal.signal(signal.SIGINT, self.__signal_cb)
+        signal.signal(signal.SIGQUIT, self.__signal_cb)
+        signal.signal(signal.SIGABRT, self.__signal_cb)
+        signal.signal(signal.SIGTERM, self.__signal_cb)
 
-        if os.path.exists(KASUMI_IMG_PATH):
-            label = u""
-            icon = unicode(KASUMI_IMG_PATH)
+    def __signal_cb(self, signum, object):
+        self.__remove_dict_files()
+        signal.signal(signum, signal.SIG_DFL)
+        os.kill(os.getpid(), signum)
+
+    def __set_segment_mode_props(self, anthy_props):
+        segment_mode_prop = ibus.Property(key=u"SegmentMode",
+                                          type=ibus.PROP_TYPE_MENU,
+                                          label=u"連",
+                                          tooltip=UN(_("Switch conversion mode")))
+        self.__prop_dict[u"SegmentMode"] = segment_mode_prop
+
+        props = ibus.PropList()
+        props.append(ibus.Property(key=u"SegmentMode.Multi",
+                                   type=ibus.PROP_TYPE_RADIO,
+                                   label=UN(_("Multiple segment"))))
+        props.append(ibus.Property(key=u"SegmentMode.Single",
+                                   type=ibus.PROP_TYPE_RADIO,
+                                   label=UN(_("Single segment"))))
+        props.append(ibus.Property(key=u"SegmentMode.ImmediateMulti",
+                                   type=ibus.PROP_TYPE_RADIO,
+                                   label=UN(_("Immediate conversion (Multiple segment)"))))
+        props.append(ibus.Property(key=u"SegmentMode.ImmediateSingle",
+                                   type=ibus.PROP_TYPE_RADIO,
+                                   label=UN(_("Immediate conversion (Single segment)"))))
+        props[self.__segment_mode].set_state(ibus.PROP_STATE_CHECKED)
+
+        for prop in props:
+            self.__prop_dict[prop.key] = prop
+
+        segment_mode_prop.set_sub_props(props)
+        anthy_props.append(segment_mode_prop)
+
+    def __set_dict_mode_props(self, anthy_props):
+        short_label = self.__prefs.get_value('dict/file/embedded',
+                                             'short_label')
+        dict_mode_prop = ibus.Property(key=u"DictMode",
+                                       type=ibus.PROP_TYPE_MENU,
+                                       label=UN(short_label),
+                                       tooltip=UN(_("Switch Dictionary")))
+        self.__prop_dict[u"DictMode"] = dict_mode_prop
+        props = ibus.PropList()
+
+        long_label = self.__prefs.get_value('dict/file/embedded',
+                                            'long_label')
+        props.append(ibus.Property(key=u"DictMode.embedded",
+                                   type=ibus.PROP_TYPE_RADIO,
+                                   label=UN(_(long_label))))
+
+        for file in self.__prefs.get_value('dict', 'files'):
+            self._link_dict_file(file)
+            id = self._get_dict_id_from_file(file)
+            if id == None:
+                continue
+            section = 'dict/file/' + id
+            if not self.__prefs.get_value(section, 'single'):
+                continue
+            key = "DictMode." + id
+            long_label = self.__prefs.get_value(section, 'long_label')
+            if 'is_system' in self.__prefs.keys(section) and \
+               self.__prefs.get_value(section, 'is_system'):
+                uni_long_label = UN(_(long_label))
+            else:
+                uni_long_label = UN(long_label)
+            props.append(ibus.Property(key=UN(key),
+                                       type=ibus.PROP_TYPE_RADIO,
+                                       label=uni_long_label))
+
+        props[self.__dict_mode].set_state(ibus.PROP_STATE_CHECKED)
+        for prop in props:
+            self.__prop_dict[prop.key] = prop
+        dict_mode_prop.set_sub_props(props)
+        anthy_props.append(dict_mode_prop)
+        self.__init_signal()
+
+    def __set_dict_config_props(self, anthy_props):
+        admin_command = self.__prefs.get_value('common', 'dict_admin_command')
+        icon_path = self.__prefs.get_value('common', 'dict_config_icon')
+
+        if not path.exists(admin_command[0]):
+            return
+        label = UN(_("Dictionary - Anthy"))
+        if icon_path and path.exists(icon_path):
+            icon = unicode(icon_path)
         else:
             # Translators: "Dic" means 'dictionary', One kanji may be good.
             label = UN(_("Dic"))
@@ -247,6 +355,22 @@ class Engine(ibus.EngineBase):
 
         return clipboard_text
 
+    def __get_single_dict_files(self):
+        files = self.__prefs.get_value('dict', 'files')
+        single_files = []
+        for file in files:
+            id = self._get_dict_id_from_file(file)
+            if id == None:
+                continue
+            section = 'dict/file/' + id
+            if self.__prefs.get_value(section, 'single'):
+                single_files.append(file)
+        return single_files
+
+    def __remove_dict_files(self):
+        for file in self.__prefs.get_value('dict', 'files'):
+            self._remove_dict_file(file)
+
     def update_preedit(self, string, attrs, cursor_pos, visible):
         self.update_preedit_text(ibus.Text(string, attrs), cursor_pos, visible)
 
@@ -426,9 +550,19 @@ class Engine(ibus.EngineBase):
     def property_activate(self, prop_name, state):
 
         if state == ibus.PROP_STATE_CHECKED:
-            if self.__input_mode_activate(prop_name, state):
+            if prop_name == None:
                 return
-            if self.__typing_mode_activate(prop_name, state):
+            elif prop_name.startswith(u"InputMode."):
+                self.__input_mode_activate(prop_name, state)
+                return
+            elif prop_name.startswith(u"TypingMode."):
+                self.__typing_mode_activate(prop_name, state)
+                return
+            elif prop_name.startswith(u"SegmentMode."):
+                self.__segment_mode_activate(prop_name, state)
+                return
+            elif prop_name.startswith(u"DictMode."):
+                self.__dict_mode_activate(prop_name, state)
                 return
         else:
             if prop_name == 'setup':
@@ -439,11 +573,14 @@ class Engine(ibus.EngineBase):
                 self.__start_add_word()
             else:
                 self.__prop_dict[prop_name].set_state(state)
+                if prop_name == "DictMode":
+                    sub_name = self.__dict_mode_get_prop_name(self.__dict_mode)
+                    if sub_name == None:
+                        return
+                    self.__dict_mode_activate(sub_name,
+                                              ibus.PROP_STATE_CHECKED)
 
     def __input_mode_activate(self, prop_name, state):
-        if not prop_name.startswith(u"InputMode."):
-            return False
-
         input_modes = {
             u"InputMode.Hiragana" : (INPUT_MODE_HIRAGANA, u"あ"),
             u"InputMode.Katakana" : (INPUT_MODE_KATAKANA, u"ア"),
@@ -454,13 +591,13 @@ class Engine(ibus.EngineBase):
 
         if prop_name not in input_modes:
             print >> sys.stderr, "Unknow prop_name = %s" % prop_name
-            return True
+            return
         self.__prop_dict[prop_name].set_state(state)
         self.update_property(self.__prop_dict[prop_name])
 
         mode, label = input_modes[prop_name]
         if self.__input_mode == mode:
-            return True
+            return
 
         self.__input_mode = mode
         prop = self.__prop_dict[u"InputMode"]
@@ -471,9 +608,6 @@ class Engine(ibus.EngineBase):
         self.__invalidate()
 
     def __typing_mode_activate(self, prop_name, state):
-        if not prop_name.startswith(u"TypingMode."):
-            return False
-
         typing_modes = {
             u"TypingMode.Romaji" : (jastring.TYPING_MODE_ROMAJI, u"R"),
             u"TypingMode.Kana" : (jastring.TYPING_MODE_KANA, u"か"),
@@ -482,9 +616,11 @@ class Engine(ibus.EngineBase):
 
         if prop_name not in typing_modes:
             print >> sys.stderr, "Unknow prop_name = %s" % prop_name
-            return True
+            return
         self.__prop_dict[prop_name].set_state(state)
         self.update_property(self.__prop_dict[prop_name])
+        if prop_name == u"TypingMode.ThumbShift":
+            self._reset_thumb()
 
         mode, label = typing_modes[prop_name]
 
@@ -512,6 +648,82 @@ class Engine(ibus.EngineBase):
         prop.label = label
         self.update_property(prop)
 
+    def __segment_mode_activate(self, prop_name, state):
+        segment_modes = {
+            u"SegmentMode.Multi" : (SEGMENT_DEFAULT, u"連"),
+            u"SegmentMode.Single" : (SEGMENT_SINGLE, u"単"),
+            u"SegmentMode.ImmediateMulti" : (SEGMENT_IMMEDIATE, u"逐|連"),
+            u"SegmentMode.ImmediateSingle" :
+                (SEGMENT_IMMEDIATE | SEGMENT_SINGLE, u"逐|単"),
+        }
+
+        if prop_name not in segment_modes:
+            print >> sys.stderr, "Unknow prop_name = %s" % prop_name
+            return
+        self.__prop_dict[prop_name].set_state(state)
+        self.update_property(self.__prop_dict[prop_name])
+
+        mode, label = segment_modes[prop_name]
+
+        self.__segment_mode = mode
+        prop = self.__prop_dict[u"SegmentMode"]
+        prop.label = label
+        self.update_property(prop)
+
+        self.__reset()
+        self.__invalidate()
+
+    def __dict_mode_get_prop_name(self, mode):
+        if mode == 0:
+            id = 'embedded'
+        else:
+            single_files = self.__get_single_dict_files()
+            file = single_files[mode - 1]
+            id = self._get_dict_id_from_file(file)
+        if id == None:
+            return None
+        return 'DictMode.' + id
+
+    def __dict_mode_activate(self, prop_name, state):
+        if prop_name not in self.__prop_dict.keys():
+            # The prop_name is added. Need to restart.
+            return
+        i = prop_name.find('.')
+        if i < 0:
+            return
+        id = prop_name[i + 1:].encode('utf-8')
+
+        file = None
+        files = self.__prefs.get_value('dict', 'files')
+        if id == 'embedded':
+            pass
+        elif id == 'anthy_zipcode' or id == 'ibus_symbol':
+            file = self.__prefs.get_value('dict', id)[0]
+        else:
+            found = False
+            for file in files:
+                if id == self._get_quoted_id(file):
+                    found = True
+                    break
+            if found == False:
+                return
+
+        if id == 'embedded':
+            dict_name = 'default'
+            self.__dict_mode = 0
+        else:
+            dict_name = 'ibus__' + id
+            self.__dict_mode = files.index(file) + 1
+        self.__prop_dict[prop_name].set_state(state)
+        self.update_property(self.__prop_dict[prop_name])
+        self.__context.init_personality()
+        self.__context.do_set_personality(dict_name)
+
+        prop = self.__prop_dict[u"DictMode"]
+        section = 'dict/file/' + id
+        prop.label = self.__prefs.get_value(section, 'short_label')
+        self.update_property(prop)
+
     def focus_in(self):
         self.register_properties(self.__prop_list)
         self.__refresh_typing_mode_property()
@@ -526,8 +738,32 @@ class Engine(ibus.EngineBase):
         elif mode == 1:
             self.__on_key_return()
 
+    def reset(self):
+        self.__reset()
+        self.__invalidate()
+
+    def do_destroy(self):
+        if self.__idle_id != 0:
+            gobject.source_remove(self.__idle_id)
+            self.__idle_id = 0
+        self.__remove_dict_files()
+        super(Engine,self).do_destroy()
+
+    def __join_all_segments(self):
+        while True:
+            conv_stat = anthy.anthy_conv_stat()
+            self.__context.get_stat(conv_stat)
+            seg = conv_stat.nr_segment - self.__cursor_pos
+
+            if seg > 1:
+                self.__context.resize_segment(self.__cursor_pos, 1)
+            else:
+                break
+
     # begine convert
     def __begin_anthy_convert(self):
+        if self.__segment_mode & SEGMENT_IMMEDIATE:
+            self.__end_anthy_convert()
         if self.__convert_mode == CONV_MODE_ANTHY:
             return
         self.__convert_mode = CONV_MODE_ANTHY
@@ -536,6 +772,8 @@ class Engine(ibus.EngineBase):
         text, cursor = self.__preedit_ja_string.get_hiragana(True)
 
         self.__context.set_string(text.encode("utf8"))
+        if self.__segment_mode & SEGMENT_SINGLE:
+            self.__join_all_segments()
         conv_stat = anthy.anthy_conv_stat()
         self.__context.get_stat(conv_stat)
 
@@ -544,7 +782,10 @@ class Engine(ibus.EngineBase):
             text = unicode(buf, "utf-8")
             self.__segments.append((0, text))
 
-        self.__cursor_pos = 0
+        if self.__segment_mode & SEGMENT_IMMEDIATE:
+            self.__cursor_pos = conv_stat.nr_segment - 1
+        else:
+            self.__cursor_pos = 0
         self.__fill_lookup_table()
         self.__lookup_table_visible = False
 
@@ -562,6 +803,52 @@ class Engine(ibus.EngineBase):
     def __end_convert(self):
         self.__end_anthy_convert()
 
+    def __candidate_cb(self, candidate):
+        if not self.__is_utf8:
+            return
+        for key in romaji_utf8_rule.keys():
+            if candidate.find(key) >= 0:
+                for value in romaji_utf8_rule[key]:
+                    candidate = candidate.replace(key, value)
+                    self.__lookup_table.append_candidate(ibus.Text(candidate))
+
+    def __fill_anthy_zipcode_strip(self, dict_file, id):
+        import re
+        text = self.__preedit_ja_string.get_latin()[0]
+        if text.find('-') < 0:
+            return
+        text = text.replace('-', '')
+        section = 'dict/file/' + id
+        if 'encoding' not in self.__prefs.keys(section):
+            section = 'dict/file/default'
+        encoding = self.__prefs.get_value(section, 'encoding')
+        contents = unicode(open(dict_file).read(), encoding)
+        expression = re.compile("^" + text + "[ \t]")
+
+        found = False
+        dict_dest = None
+        for line in contents.split('\n'):
+            matched = expression.search(line)
+            if matched:
+                found = True
+                dict_dest = unicode(matched.string).split(' ')[2]
+                break
+        if found:
+            self.__lookup_table.append_candidate(ibus.Text(dict_dest))
+
+    def __fill_lookup_table_dict_mode(self):
+        if self.__dict_mode <= 0:
+            return
+        single_files = self.__get_single_dict_files()
+        file = single_files[self.__dict_mode - 1]
+        if file == None:
+            return
+        id = self._get_dict_id_from_file(file)
+        if id == None:
+            return
+        if id == 'anthy_zipcode':
+            self.__fill_anthy_zipcode_strip(file, id)
+
     def __fill_lookup_table(self):
         if self.__convert_mode == CONV_MODE_PREDICTION:
             seg_stat = anthy.anthy_prediction_stat()
@@ -573,7 +860,7 @@ class Engine(ibus.EngineBase):
                 buf = self.__context.get_prediction(i)
                 candidate = unicode(buf, "utf-8")
                 self.__lookup_table.append_candidate(ibus.Text(candidate))
-
+                self.__candidate_cb(candidate)
             return
 
         # get segment stat
@@ -586,13 +873,15 @@ class Engine(ibus.EngineBase):
             buf = self.__context.get_segment(self.__cursor_pos, i)
             candidate = unicode(buf, "utf-8")
             self.__lookup_table.append_candidate(ibus.Text(candidate))
+            self.__candidate_cb(candidate)
+        self.__fill_lookup_table_dict_mode()
 
 
     def __invalidate(self):
-        if self.__need_update:
+        if self.__idle_id != 0:
             return
-        self.__need_update = True
-        gobject.idle_add(self.__update, priority = gobject.PRIORITY_LOW)
+        self.__idle_id = gobject.idle_add(self.__update,
+                                          priority = gobject.PRIORITY_LOW)
 
 #    def __get_preedit(self):
     def __get_preedit(self, commit=False):
@@ -699,11 +988,11 @@ class Engine(ibus.EngineBase):
             self.__lookup_table_visible)
 
     def __update(self):
-        self.__need_update = False
         if self.__convert_mode == CONV_MODE_OFF:
             self.__update_input_chars()
         else:
             self.__update_convert_chars()
+        self.__idle_id = 0
 
     def __on_key_return(self):
         if self.__preedit_ja_string.is_empty():
@@ -956,7 +1245,7 @@ class Engine(ibus.EngineBase):
         self.__invalidate()
         return True
 
-    def __on_key_common(self, keyval):
+    def __on_key_common(self, keyval, state=0):
 
         if self.__input_mode == INPUT_MODE_LATIN:
             # Input Latin chars
@@ -974,14 +1263,26 @@ class Engine(ibus.EngineBase):
             return True
 
         # Input Japanese
-        if self.__convert_mode == CONV_MODE_ANTHY:
+        if self.__segment_mode & SEGMENT_IMMEDIATE:
+            # Commit nothing
+            pass
+        elif self.__convert_mode == CONV_MODE_ANTHY:
             for i, (seg_index, text) in enumerate(self.__segments):
                 self.__context.commit_segment(i, seg_index)
             self.__commit_string(self.__convert_chars)
         elif self.__convert_mode != CONV_MODE_OFF:
             self.__commit_string(self.__convert_chars)
 
+        # "n" + "'" == "nn" in romaji
+        if (keyval >= ord('A') and keyval <= ord('Z')) or \
+           (keyval >= ord('a') and keyval <= ord('z')):
+            shift = (state & modifier.SHIFT_MASK) != 0
+        else:
+            shift = False
+        self.__preedit_ja_string.set_shift(shift)
         self.__preedit_ja_string.insert(unichr(keyval))
+        if self.__segment_mode & SEGMENT_IMMEDIATE:
+            self.__begin_anthy_convert()
         self.__invalidate()
         return True
 
@@ -1024,6 +1325,15 @@ class Engine(ibus.EngineBase):
             cls.__prefs.set_value(base_sec, name, value)
             if name == 'shortcut_type':
                 cls.__keybind = cls._mk_keybind()
+        elif base_sec == 'thumb':
+            cls.__prefs.set_value(base_sec, name, value)
+            cls._reset_thumb()
+        elif base_sec == 'dict':
+            cls._set_dict_files_value(base_sec, name, value)
+        elif base_sec.startswith('dict/file/'):
+            if base_sec not in cls.__prefs.sections():
+                cls._fetch_dict_values(base_sec)
+            cls.__prefs.set_value(base_sec, name, value)
         elif base_sec:
             cls.__prefs.set_value(base_sec, name, value)
         else:
@@ -1056,6 +1366,175 @@ class Engine(ibus.EngineBase):
                  'alt+' in s and modifier.MOD1_MASK or 0)
         return cls._mk_key(keyval, state)
 
+    @classmethod
+    def _reset_thumb(cls):
+        if cls.__thumb == None:
+            import thumb
+            cls.__thumb = thumb.ThumbShiftKeyboard(cls.__prefs)
+
+        else:
+            cls.__thumb.reset()
+
+    @classmethod
+    def _get_userhome(cls):
+        if 'HOME' not in environ:
+            import pwd
+            userhome = pwd.getpwuid(getuid()).pw_dir
+        else:
+            userhome = environ['HOME']
+        userhome = userhome.rstrip('/')
+        return userhome
+
+    @classmethod
+    def _get_quoted_id(cls, file):
+        id = file
+        has_mbcs = False
+
+        for i in xrange(0, len(id)):
+            if ord(id[i]) >= 0x7f:
+                    has_mbcs = True
+                    break
+        if has_mbcs:
+            import urllib
+            id = urllib.quote(id)
+
+        if id.find('/') >=0:
+            id = id[id.rindex('/') + 1:]
+        if id.find('.') >=0:
+            id = id[:id.rindex('.')]
+        return id
+
+    @classmethod
+    def _get_dict_id_from_file(cls, file):
+        if file in cls.__prefs.get_value('dict', 'anthy_zipcode'):
+            id = 'anthy_zipcode'
+        elif file in cls.__prefs.get_value('dict', 'ibus_symbol'):
+            id = 'ibus_symbol'
+        else:
+            id = cls._get_quoted_id(file)
+        return id
+
+    @classmethod
+    def _link_dict_file_with_id(cls, file, id, link_mode):
+        if not path.exists(file):
+            print >> sys.stderr, file + " does not exist"
+            return
+        if id == None:
+            return
+        if link_mode == LINK_DICT_EMBEDDED:
+            directory = cls._get_userhome() + "/.anthy/" + IMPORTED_EMBEDDED_DICT_DIR
+            name = IMPORTED_EMBEDDED_DICT_PREFIX + id
+        elif link_mode == LINK_DICT_SINGLE:
+            directory = cls._get_userhome() + "/.anthy"
+            name = IMPORTED_SINGLE_DICT_PREFIX + id
+        else:
+            return
+        if path.exists(directory):
+            if not path.isdir(directory):
+                print >> sys.stderr, directory + " is not a directory"
+                return
+        else:
+            os.makedirs(directory, 0700)
+        backup_dir = os.getcwd()
+        os.chdir(directory)
+        if path.lexists(directory + '/' + name):
+            if path.islink(directory + '/' + name):
+                print >> sys.stderr, "Removing " + name
+                os.unlink(directory + '/' + name)
+            else:
+                alternate = name + str(os.getpid())
+                print >> sys.stderr, "Moving " + name + " to " + alternate
+                os.rename(name, alternate)
+        os.symlink(file, directory + '/' + name)
+        if backup_dir != None:
+            os.chdir(backup_dir)
+
+    @classmethod
+    def _remove_dict_file_with_id(cls, file, id, link_mode):
+        if id == None:
+            return
+        if link_mode == LINK_DICT_EMBEDDED:
+            directory = cls._get_userhome() + "/.anthy/" + IMPORTED_EMBEDDED_DICT_DIR
+            name = IMPORTED_EMBEDDED_DICT_PREFIX + id
+        elif link_mode == LINK_DICT_SINGLE:
+            directory = cls._get_userhome() + "/.anthy"
+            name = IMPORTED_SINGLE_DICT_PREFIX + id
+        else:
+            return
+        if path.exists(directory):
+            if not path.isdir(directory):
+                print >> sys.stderr, directory + " is not a directory"
+                return
+        backup_dir = os.getcwd()
+        os.chdir(directory)
+        if path.lexists(directory + '/' + name):
+            os.unlink(directory + '/' + name)
+        if backup_dir != None:
+            os.chdir(backup_dir)
+
+    @classmethod
+    def _link_dict_file(cls, file):
+        id = cls._get_dict_id_from_file(file)
+        if id == None:
+            return
+        section = 'dict/file/' + id
+        if section not in cls.__prefs.sections():
+            cls._fetch_dict_values(section)
+        if cls.__prefs.get_value(section, 'embed'):
+            cls._link_dict_file_with_id(file, id, LINK_DICT_EMBEDDED)
+        if cls.__prefs.get_value(section, 'single'):
+            cls._link_dict_file_with_id(file, id, LINK_DICT_SINGLE)
+
+    @classmethod
+    def _remove_dict_file(cls, file):
+        id = cls._get_dict_id_from_file(file)
+        if id == None:
+            return
+        section = 'dict/file/' + id
+        if section not in cls.__prefs.sections():
+            cls._fetch_dict_values(section)
+        if cls.__prefs.get_value(section, 'embed'):
+            cls._remove_dict_file_with_id(file, id, LINK_DICT_EMBEDDED)
+        if cls.__prefs.get_value(section, 'single'):
+            cls._remove_dict_file_with_id(file, id, LINK_DICT_SINGLE)
+
+    @classmethod
+    def _set_dict_files_value(cls, base_sec, name, value):
+        if name == 'files':
+            str_list = []
+            for file in value:
+                str_list.append(str(file))
+            old_files = cls.__prefs.get_value(base_sec, name)
+            for file in old_files:
+                if file in str_list:
+                    continue
+                cls._remove_dict_file(file)
+            for file in str_list:
+                if file in old_files:
+                    continue
+                cls._link_dict_file(file)
+            cls.__prefs.set_value(base_sec, name, str_list)
+        else:
+            cls.__prefs.set_value(base_sec, name, value)
+
+    @classmethod
+    def _fetch_dict_values(cls, section):
+        cls.__prefs.set_new_section(section)
+        cls.__prefs.set_new_key(section, 'short_label')
+        cls.__prefs.fetch_item(section, 'short_label')
+        cls.__prefs.set_value(section, 'short_label',
+                              str(cls.__prefs.get_value(section, 'short_label')))
+        cls.__prefs.set_new_key(section, 'long_label')
+        cls.__prefs.fetch_item(section, 'long_label')
+        cls.__prefs.set_value(section, 'long_label',
+                              str(cls.__prefs.get_value(section, 'long_label')))
+        cls.__prefs.set_new_key(section, 'embed')
+        cls.__prefs.fetch_item(section, 'embed')
+        cls.__prefs.set_new_key(section, 'single')
+        cls.__prefs.fetch_item(section, 'single')
+        cls.__prefs.set_new_key(section, 'reverse')
+        cls.__prefs.fetch_item(section, 'reverse')
+
     @staticmethod
     def _mk_key(keyval, state):
         if state & (modifier.CONTROL_MASK | modifier.MOD1_MASK):
@@ -1067,12 +1546,12 @@ class Engine(ibus.EngineBase):
         return repr([int(state), int(keyval)])
 
     def process_key_event_thumb(self, keyval, keycode, state):
-        import gtk
-        import thumb
+        if self.__thumb == None:
+            self._reset_thumb()
 
         def on_timeout(keyval):
             if self._MM:
-                insert(thumb.table[self._MM][self._SS])
+                insert(self.__thumb.get_char(self._MM)[self._SS])
             else:
                 cmd_exec([0, RS(), LS()][self._SS])
             self._H = None
@@ -1089,6 +1568,7 @@ class Engine(ibus.EngineBase):
 
         def insert(keyval):
             try:
+                self._MM = self._SS = 0
                 ret = self.__on_key_common(ord(keyval))
                 if (keyval in u',.、。' and
                     self.__prefs.get_value('common', 'behavior_on_period')):
@@ -1109,16 +1589,16 @@ class Engine(ibus.EngineBase):
             return False
 
         def RS():
-            return self.__prefs.get_value('common', 'thumb_rs')
+            return self.__thumb.get_rs()
 
         def LS():
-            return self.__prefs.get_value('common', 'thumb_ls')
+            return self.__thumb.get_ls()
 
         def T1():
-            return self.__prefs.get_value('common', 'thumb_t1')
+            return self.__thumb.get_t1()
 
         def T2():
-            return self.__prefs.get_value('common', 'thumb_t2')
+            return self.__thumb.get_t2()
 
         state = state & (modifier.SHIFT_MASK |
                          modifier.CONTROL_MASK |
@@ -1132,12 +1612,16 @@ class Engine(ibus.EngineBase):
         if state & modifier.RELEASE_MASK:
             if keyval == self._MM:
                 if stop():
-                    insert(thumb.table[self._MM][self._SS])
+                    insert(self.__thumb.get_char(self._MM)[self._SS])
                 self._MM = 0
             elif (1 if keyval == RS() else 2) == self._SS:
                 if stop():
                     cmd_exec([0, RS(), LS()][self._SS])
                 self._SS = 0
+            if keyval in [RS(), LS()]:
+                self._RSS = 0
+            elif keyval == self._RMM:
+                self._RMM = 0
         else:
             if keyval in [LS(), RS()] and state == 0:
                 if self._SS:
@@ -1147,35 +1631,50 @@ class Engine(ibus.EngineBase):
                     start(T1())
                 elif self._MM:
                     stop()
-                    insert(thumb.table[self._MM][1 if keyval == RS() else 2])
+                    self._RMM = self._MM
+                    self._RSS = 1 if keyval == RS() else 2
+                    insert(self.__thumb.get_char(self._MM)[1 if keyval == RS() else 2])
                 else:
-                    self._SS = 1 if keyval == RS() else 2
-                    start(T1())
-            elif keyval in thumb.table.keys() and state == 0:
+                    if self._RSS == (1 if keyval == RS() else 2):
+                        if self._RMM:
+                            insert(self.__thumb.get_char(self._RMM)[self._RSS])
+                    else:
+                        self._SS = 1 if keyval == RS() else 2
+                        start(T1())
+            elif keyval in self.__thumb.get_chars() and state == 0:
                 if self._MM:
                     stop()
-                    insert(thumb.table[self._MM][self._SS])
+                    insert(self.__thumb.get_char(self._MM)[self._SS])
                     start(T2())
                     self._MM = keyval
                 elif self._SS:
                     stop()
-                    insert(thumb.table[keyval][self._SS])
+                    self._RMM = keyval
+                    self._RSS = self._SS
+                    insert(self.__thumb.get_char(keyval)[self._SS])
                 else:
-                    if cmd_exec(keyval, state):
-                        return True
-                    start(T2())
-                    self._MM = keyval
+                    if self._RMM  == keyval:
+                        if self._RSS:
+                            insert(self.__thumb.get_char(self._RMM)[self._RSS])
+                    else:
+                        if cmd_exec(keyval, state):
+                            return True
+                        start(T2())
+                        self._MM = keyval
             else:
                 if self._MM:
                     stop()
-                    insert(thumb.table[self._MM][self._SS])
+                    insert(self.__thumb.get_char(self._MM)[self._SS])
                 elif self._SS:
                     stop()
                     cmd_exec([0, RS(), LS()][self._SS])
                 if cmd_exec(keyval, state):
                     return True
                 elif 0x21 <= keyval <= 0x7e and state & (modifier.CONTROL_MASK | modifier.MOD1_MASK) == 0:
-                    insert(thumb.shift_table.get(keyval, unichr(keyval)))
+                    if state & modifier.SHIFT_MASK:
+                        insert(self.__thumb.get_shift_char(keyval, unichr(keyval)))
+                    elif self._SS == 0:
+                        insert(unichr(keyval))
                 else:
                     if not self.__preedit_ja_string.is_empty():
                         return True
@@ -1220,7 +1719,7 @@ class Engine(ibus.EngineBase):
                     keyval = keysyms.asciitilde
                 elif keyval == keysyms.backslash and keycode in [132-8, 133-8]:
                     keyval = keysyms.yen
-            ret = self.__on_key_common(keyval)
+            ret = self.__on_key_common(keyval, state)
             if (unichr(keyval) in u',.' and
                 self.__prefs.get_value('common', 'behavior_on_period')):
                 return self.__cmd_convert(keyval, state)
@@ -1318,6 +1817,22 @@ class Engine(ibus.EngineBase):
                                     ibus.PROP_STATE_CHECKED)
         return True
 
+    def __cmd_circle_dict_method(self, keyval, state):
+        if not self._chk_mode('0'):
+            return False
+
+        single_files = self.__get_single_dict_files()
+        new_mode = self.__dict_mode + 1
+        if new_mode > len(single_files):
+            new_mode = 0
+        self.__dict_mode = new_mode
+        prop_name = self.__dict_mode_get_prop_name(self.__dict_mode)
+        if prop_name == None:
+            return False
+        self.__dict_mode_activate(prop_name,
+                                  ibus.PROP_STATE_CHECKED)
+        return True
+
     #edit_keys
     def __cmd_insert_space(self, keyval, state):
         if (self.__prefs.get_value('common', 'half_width_space') or
@@ -1540,6 +2055,16 @@ class Engine(ibus.EngineBase):
             self.__shrink_segment(1)
             return True
 
+    def __move_cursor_char_length(self, length):
+        if self.__input_mode == INPUT_MODE_HIRAGANA:
+            self.__preedit_ja_string.move_cursor_hiragana_length(length)
+        elif self.__input_mode == INPUT_MODE_KATAKANA:
+            self.__preedit_ja_string.move_cursor_katakana_length(length)
+        elif self.__input_mode == INPUT_MODE_HALF_WIDTH_KATAKANA:
+            self.__preedit_ja_string.move_cursor_half_with_katakana_length(length)
+        else:
+            self.__preedit_ja_string.move_cursor(length)
+
     def __commit_nth_segment(self, commit_index, keyval, state):
 
         if commit_index >= len(self.__segments):
@@ -1555,10 +2080,10 @@ class Engine(ibus.EngineBase):
             for i in xrange(0, commit_index + 1):
                 buf = self.__context.get_segment(i, NTH_UNCONVERTED_CANDIDATE)
                 commit_length += len(unicode(buf, "utf-8"))
-            self.__preedit_ja_string.move_cursor(commit_length - cursor)
+            self.__move_cursor_char_length(commit_length - cursor)
             for i in xrange(0, commit_length):
                 self.__preedit_ja_string.remove_before()
-            self.__preedit_ja_string.move_cursor(cursor - commit_length)
+            self.__move_cursor_char_length(cursor - commit_length)
 
             del self.__segments[0:commit_index + 1]
 
@@ -1569,7 +2094,7 @@ class Engine(ibus.EngineBase):
                 self.__cursor_pos -= (commit_index + 1)
             else:
                 self.__cursor_pos = 0
-            (seg_index, text) = self.__segments[self.__cursor_pos]
+            text, cursor = self.__get_preedit()
             self.__convert_chars = text
             self.__context.set_string(text.encode ("utf-8"))
 
@@ -1844,12 +2369,12 @@ class Engine(ibus.EngineBase):
         return True
 
     def __start_dict_admin(self):
-        path = self.__prefs.get_value('common', 'dict_admin_command')
-        os.spawnl(os.P_NOWAIT, *path)
+        command = self.__prefs.get_value('common', 'dict_admin_command')
+        os.spawnl(os.P_NOWAIT, *command)
 
     def __start_add_word(self):
-        path = self.__prefs.get_value('common', 'add_word_command')
-        os.spawnl(os.P_NOWAIT, *path)
+        command = self.__prefs.get_value('common', 'add_word_command')
+        os.spawnl(os.P_NOWAIT, *command)
 
     def __start_setup(self):
         if Engine.__setup_pid != 0:
